@@ -31,18 +31,20 @@
 #include "../../kexec.h"
 #include "../../kexec-syscall.h"
 #include "kexec-ppc64.h"
+#include "../../fs2dt.h"
 #include "crashdump-ppc64.h"
 #include <arch/options.h>
 
 static struct memory_range *exclude_range = NULL;
 static struct memory_range *memory_range = NULL;
 static struct memory_range *base_memory_range = NULL;
-static uint64_t rmo_top;
+static uint64_t rma_top;
 uint64_t memory_max = 0;
 uint64_t memory_limit;
 static int nr_memory_ranges, nr_exclude_ranges;
 uint64_t crash_base, crash_size;
 unsigned int rtas_base, rtas_size;
+uint64_t opal_base, opal_size;
 int max_memory_ranges;
 
 static void cleanup_memory_ranges(void)
@@ -313,12 +315,77 @@ static int sort_ranges(void)
 	return 0;
 }
 
+void scan_reserved_ranges(unsigned long kexec_flags, int *range_index)
+{
+	char fname[256], buf[16];
+	FILE *file;
+	int i = *range_index;
+
+	strcpy(fname, "/proc/device-tree/reserved-ranges");
+
+	file = fopen(fname, "r");
+	if (file == NULL) {
+		if (errno != ENOENT) {
+			perror(fname);
+			return;
+		}
+		errno = 0;
+		/* File not present. Non PowerKVM system. */
+		return;
+	}
+
+	/*
+	 * Each reserved range is an (address,size) pair, 2 cells each,
+	 * totalling 4 cells per range.
+	 */
+	while (fread(buf, sizeof(uint64_t) * 2, 1, file) == 1) {
+		uint64_t base, size;
+
+		base = be64_to_cpu(((uint64_t *)buf)[0]);
+		size = be64_to_cpu(((uint64_t *)buf)[1]);
+
+		exclude_range[i].start = base;
+		exclude_range[i].end = base + size;
+		i++;
+		if (i >= max_memory_ranges)
+			realloc_memory_ranges();
+
+		reserve(base, size);
+	}
+	fclose(file);
+	*range_index = i;
+}
+
+/* Return 0 if fname/value valid, -1 otherwise */
+int get_devtree_value(const char *fname, unsigned long long *value)
+{
+	FILE *file;
+	char buf[MAXBYTES];
+	int n = -1;
+
+	if ((file = fopen(fname, "r"))) {
+		n = fread(buf, 1, MAXBYTES, file);
+		fclose(file);
+	}
+
+	if (n == sizeof(uint32_t))
+		*value = ((uint32_t *)buf)[0];
+	else if (n == sizeof(uint64_t))
+		*value = ((uint64_t *)buf)[0];
+	else {
+		fprintf(stderr, "%s node has invalid size: %d\n", fname, n);
+		return -1;
+	}
+
+	return 0;
+}
+
 /* Get devtree details and create exclude_range array
  * Also create usablemem_ranges for KEXEC_ON_CRASH
  */
 static int get_devtree_details(unsigned long kexec_flags)
 {
-	uint64_t rmo_base;
+	uint64_t rma_base = -1, base;
 	uint64_t tce_base;
 	unsigned int tce_size;
 	uint64_t htab_base, htab_size;
@@ -338,12 +405,15 @@ static int get_devtree_details(unsigned long kexec_flags)
 		return -1;
 	}
 
+	scan_reserved_ranges(kexec_flags, &i);
+
 	while ((dentry = readdir(dir)) != NULL) {
 		if (strncmp(dentry->d_name, "chosen", 6) &&
 			strncmp(dentry->d_name, "memory@", 7) &&
 			strcmp(dentry->d_name, "memory") &&
 			strncmp(dentry->d_name, "pci@", 4) &&
-			strncmp(dentry->d_name, "rtas", 4)) 
+			strncmp(dentry->d_name, "rtas", 4) &&
+			strncmp(dentry->d_name, "ibm,opal", 8))
 			continue;
 		strcpy(fname, device_tree);
 		strcat(fname, dentry->d_name);
@@ -575,6 +645,46 @@ static int get_devtree_details(unsigned long kexec_flags)
 				add_usable_mem_rgns(rtas_base, rtas_size);
 		} /* rtas */
 
+		if (strncmp(dentry->d_name, "ibm,opal", 8) == 0) {
+			strcat(fname, "/opal-base-address");
+			file = fopen(fname, "r");
+			if (file == NULL) {
+				perror(fname);
+				goto error_opencdir;
+			}
+			if (fread(&opal_base, sizeof(uint64_t), 1, file) != 1) {
+				perror(fname);
+				goto error_openfile;
+			}
+			opal_base = be64_to_cpu(opal_base);
+			fclose(file);
+
+			memset(fname, 0, sizeof(fname));
+			strcpy(fname, device_tree);
+			strcat(fname, dentry->d_name);
+			strcat(fname, "/opal-runtime-size");
+			file = fopen(fname, "r");
+			if (file == NULL) {
+				perror(fname);
+				goto error_opencdir;
+			}
+			if (fread(&opal_size, sizeof(uint64_t), 1, file) != 1) {
+				perror(fname);
+				goto error_openfile;
+			}
+			fclose(file);
+			closedir(cdir);
+			opal_size = be64_to_cpu(opal_size);
+			/* Add OPAL to exclude_range */
+			exclude_range[i].start = opal_base;
+			exclude_range[i].end = opal_base + opal_size;
+			i++;
+			if (i >= max_memory_ranges)
+				realloc_memory_ranges();
+			if (kexec_flags & KEXEC_ON_CRASH)
+				add_usable_mem_rgns(opal_base, opal_size);
+		} /* ibm,opal */
+
 		if (!strncmp(dentry->d_name, "memory@", 7) ||
 			!strcmp(dentry->d_name, "memory")) {
 			strcat(fname, "/reg");
@@ -586,10 +696,13 @@ static int get_devtree_details(unsigned long kexec_flags)
 				perror(fname);
 				goto error_openfile;
 			}
-			rmo_base = be64_to_cpu(((uint64_t *)buf)[0]);
-			rmo_top = rmo_base + be64_to_cpu(((uint64_t *)buf)[1]);
-			if (rmo_top > 0x30000000UL)
-				rmo_top = 0x30000000UL;
+			base = be64_to_cpu(((uint64_t *)buf)[0]);
+			if (base < rma_base) {
+				rma_base = base;
+				rma_top = base + be64_to_cpu(((uint64_t *)buf)[1]);
+				if (rma_top > 0x30000000UL)
+					rma_top = 0x30000000UL;
+			}
 
 			fclose(file);
 			closedir(cdir);
@@ -701,14 +814,14 @@ int setup_memory_ranges(unsigned long kexec_flags)
 				j++;
 				if (j >= max_memory_ranges)
 					realloc_memory_ranges();
-				/* Limit the end to rmo_top */
-				if (memory_range[j-1].start >= rmo_top) {
+				/* Limit the end to rma_top */
+				if (memory_range[j-1].start >= rma_top) {
 					j--;
 					break;
 				}
-				if ((memory_range[j-1].start < rmo_top) &&
-				(memory_range[j-1].end >= rmo_top)) {
-					memory_range[j-1].end = rmo_top;
+				if ((memory_range[j-1].start < rma_top) &&
+				(memory_range[j-1].end >= rma_top)) {
+					memory_range[j-1].end = rma_top;
 					break;
 				}
 				continue;
@@ -723,14 +836,14 @@ int setup_memory_ranges(unsigned long kexec_flags)
 		j++;
 		if (j >= max_memory_ranges)
 			realloc_memory_ranges();
-		/* Limit range to rmo_top */
-		if (memory_range[j-1].start >= rmo_top) {
+		/* Limit range to rma_top */
+		if (memory_range[j-1].start >= rma_top) {
 			j--;
 			break;
 		}
-		if ((memory_range[j-1].start < rmo_top) &&
-			(memory_range[j-1].end >= rmo_top)) {
-			memory_range[j-1].end = rmo_top;
+		if ((memory_range[j-1].start < rma_top) &&
+			(memory_range[j-1].end >= rma_top)) {
+			memory_range[j-1].end = rma_top;
 			break;
 		}
 	}
@@ -783,6 +896,8 @@ int file_types = sizeof(file_type) / sizeof(file_type[0]);
 void arch_usage(void)
 {
 	fprintf(stderr, "     --elf64-core-headers Prepare core headers in ELF64 format\n");
+	fprintf(stderr, "     --dt-no-old-root Do not reuse old kernel root= param.\n" \
+	                "                      while creating flatten device tree.\n");
 }
 
 struct arch_options_t arch_options = {
@@ -809,6 +924,9 @@ int arch_process_options(int argc, char **argv)
 		case OPT_ELF64_CORE:
 			arch_options.core_header_type = CORE_TYPE_ELF64;
 			break;
+		case OPT_DT_NO_OLD_ROOT:
+			dt_no_old_root = 1;
+			break;
 		}
 	}
 	/* Reset getopt for the next pass; called in other source modules */
